//
// bigdec.hpp
//
// Bill Seymour, 2022-11-13
//
// Copyright Bill Seymour 2021, 2022.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This is the header file for a big decimal type,
// either fixed-point or floating-point, implemented
// as a sign-magnitude, little-endian deque<signed char>
// that holds binary coded decimal nybbles.
//
// Non-inline functions are defined in bigdec.cpp.
//

#ifndef BIGDEC_HPP_INCLUDED
#define BIGDEC_HPP_INCLUDED

#ifdef BIGNUM_NO_IO_OPERATORS
  #ifndef BIGDEC_NO_IO_OPERATORS
    #define BIGDEC_NO_IO_OPERATORS
  #endif
  #ifndef BIGINT_NO_IO_OPERATORS
    #define BIGINT_NO_IO_OPERATORS
  #endif
#endif
#include "bigint.hpp"  // already #includes everything we need except:
#include <deque>
#include <memory> // move

#ifndef BIGDEC_MAX_DIGS
  #define BIGDEC_MAX_DIGS 38
#endif

namespace dbacc {

enum class rounding // after ISO/IEC JTC1 SC22 WG21 P1889R1 sect. 3.3
{
    all_to_neg_inf, all_to_pos_inf,
    all_to_zero, all_away_zero,
    all_to_even, all_to_odd,
    all_fastest, all_smallest,
    all_unspecified,
    tie_to_neg_inf, tie_to_pos_inf,
    tie_to_zero, tie_away_zero,
    tie_to_even, tie_to_odd,
    tie_fastest, tie_smallest,
    tie_unspecified
};

class bigdec final
{
public:
    using scale_type = int;
    static constexpr scale_type floating_scale = INT_MIN;
    static constexpr rounding default_rounding = rounding::tie_to_even;
    static constexpr rounding bankers_rounding = rounding::tie_away_zero;
    static constexpr int max_digs = BIGDEC_MAX_DIGS;

private:
//
// In this prototype, the absolute value is just a deque<signed char>
// with each element holding a single BCD digit.
//
    using elem_type = signed char;
    using absval    = std::deque<elem_type>;
    using iter      = absval::iterator;
    using citer     = absval::const_iterator;
    using riter     = absval::reverse_iterator;
    using criter    = absval::const_reverse_iterator;

    absval val;
    bool neg = false;
    scale_type this_scale = 0, user_scale;
    rounding rnd;

    void trim_trailing_zeros() noexcept
    {
        // NB:  little-endian
        while (!val.empty() && 0 == val.front())
        {
            val.pop_front();
            --this_scale;
        }
        if (val.empty())
        {
            neg = false;
            this_scale = 0;
        }
    }
    void trim_all_zeros() noexcept
    {
        while (!val.empty() && 0 == val.back())
        {
            val.pop_back();
        }
        trim_trailing_zeros();
    }

public:
//
// Special member functions, swap() and clear():
//
    bigdec(rounding r = default_rounding, scale_type s = 0)
      : user_scale(s), rnd(r) { }

    bigdec(const bigdec&) = default;
    bigdec(bigdec&&) = default;

    bigdec& operator=(const bigdec&) = default;
    bigdec& operator=(bigdec&&) = default;

    ~bigdec() noexcept { }

    void swap(bigdec& other)
    {
        val.swap(other.val);
        std::swap(neg, other.neg);
        std::swap(this_scale, other.this_scale);
        std::swap(user_scale, other.user_scale);
        std::swap(rnd, other.rnd);
    }

    void clear() noexcept
    {
        val.clear();
        neg = false;
        this_scale = 0;
        // but don't change user's requested rounding mode or scale
    }
    void all_clear() noexcept
    {
        clear();
        user_scale = 0;
        rnd = default_rounding;
    }

//
// Implicit conversions from fundamental arithmetic types other than bool:
//
private:
    void init(long long);
    void init(unsigned long long);
    void init(long double);
    #include "bignum_type_traits.inc" // is_sgn, is_uns, is_flt, is_ari
public:
    template<typename T>
    bigdec(T value,
           scale_type s = 0,
           rounding r = default_rounding,
           typename std::enable_if<is_sgn<T>::value,int>::type = 0)
      : user_scale(s), rnd(r)
    {
        init(static_cast<long long>(value));
    }
    template<typename T>
    bigdec(T value,
           scale_type s = 0,
           rounding r = default_rounding,
           typename std::enable_if<is_uns<T>::value,int>::type = 0)
      : user_scale(s), rnd(r)
    {
        init(static_cast<unsigned long long>(value));
    }
    template<typename T>
    bigdec(T value,
           scale_type s = floating_scale,
           rounding r = default_rounding,
           typename std::enable_if<is_flt<T>::value,int>::type = 0)
      : user_scale(s), rnd(r)
    {
        init(static_cast<long double>(value));
    }

//
// Our is_uns<> type trait excludes bool,
// which we probably want to be explicit anyway.
//
    explicit bigdec(bool value,
                    scale_type s = 0,
                    rounding r = rounding::all_to_zero)
      : user_scale(s), rnd(r)
    {
        if (value)
        {
            val.assign(1, 1);
        }
    }

//
// We might want to construct a bigdec from a bigint.
//
    explicit bigdec(const bigint&, scale_type = 0, rounding = default_rounding);

//
// Explicit conversions from C-style strings and std::strings:
//
private:
    void init(const char*, const char**);
public:
    explicit bigdec(const char* value,
                    rounding r = default_rounding,
                    const char** termptr = nullptr)
      : user_scale(0), rnd(r)
    {
        init(value, termptr); // infers the scale from the string
    }
    explicit bigdec(const std::string& value,
                    rounding r = default_rounding,
                    std::size_t* termpos = nullptr)
      : user_scale(0), rnd(r)
    {
        const char* beg = value.c_str();
        const char* end = beg;
        init(beg, &end);
        if (termpos != nullptr)
        {
            *termpos = static_cast<std::size_t>(end - beg);
        }
    }

//
// Assignment:
//
    template<typename T>
    typename std::enable_if<is_ari<T>::value,bigdec>&
    operator=(T value)
    {
        return *this = bigdec(value);
    }
    template<typename T>
    typename std::enable_if<is_ari<T>::value,bigdec>&
    assign(T value, scale_type s = 0, rounding r = default_rounding)
    {
        return *this = bigdec(value, s, r);
    }

    bigdec& assign(bool value,
                   scale_type s = 0,
                   rounding r = rounding::all_to_zero)
    {
        return *this = bigdec(value, s, r);
    }

    bigdec& assign(const bigint& value,
                   scale_type s = 0,
                   rounding r = default_rounding)
    {
        return *this = bigdec(value, s, r);
    }

    bigdec& assign(const char* value,
                   rounding r = default_rounding,
                   const char** termptr = nullptr)
    {
        return *this = bigdec(value, r, termptr);
    }
    bigdec& assign(const std::string& value,
                   rounding r = default_rounding,
                   std::size_t* termpos = nullptr)
    {
        return *this = bigdec(value, r, termpos);
    }

//
// Getters and setters for the scale and the rounding mode:
//
    scale_type scale() const noexcept
    {
        return user_scale;
    }
    scale_type current_scale() const noexcept
    {
        return this_scale;
    }
    bool isfloat() const noexcept
    {
        return user_scale == floating_scale;
    }

    void set_user_scale(scale_type s) noexcept
    {
        user_scale = s;
    }
    void adjust_scale_by(scale_type i) noexcept
    {
        this_scale += i;
    }
    void change_scale_to(scale_type s) noexcept
    {
        this_scale = s;
    }

    rounding get_rounding() const noexcept
    {
        return rnd;
    }
    void set_rounding(rounding r) noexcept
    {
        rnd = r;
    }

//
// Observers:
//
    bool isneg() const noexcept
    {
        return neg;
    }
    bool ispos() const noexcept
    {
        return !val.empty() && !isneg();
    }
    bool iszero() const noexcept
    {
        return val.empty();
    }
    bool isone() const noexcept
    {
        return val.size() == 1 && val[0] == 1;
    }
    bool isone(bool n) const noexcept
    {
        return neg == n && isone();
    }

    bigdec rawval() const;

    int signum() const noexcept;

//
// Conversions to other types:
//
    explicit operator bool() const noexcept
    {
        return !iszero();
    }

    explicit operator long double() const
    {
        return std::strtold(to_string().c_str(), nullptr);
    }

    bigint to_bigint(rounding = rounding::all_to_zero) const;

    std::string to_string(int, bool sci = false) const;
    std::string to_string(bool sci = false) const
    {
        int digs = isfloat() ? max_digs : val.size() + user_scale - this_scale;
        if (digs > max_digs)
        {
            digs = max_digs;
        }
        return std::move(to_string(digs, sci));
    }

//
// Mutators:
//
    bigdec& set_to_zero() noexcept
    {
        clear();
        return *this;
    }
    bigdec& set_to_one(bool n)
    {
        val.assign(1, 1);
        neg = n;
        this_scale = 0;
        return *this;
    }

    bigdec& negate() noexcept
    {
        if (!iszero())
        {
            neg = !neg;
        }
        return *this;
    }

    bigdec& abs() noexcept
    {
        neg = false;
        return *this;
    }

private:
    void inc();
    void do_round(int, rounding);
public:
    bigdec& round(int digs, rounding r)
    {
        do_round(digs, r);
        return *this;
    }
    bigdec& round(int digs)
    {
        do_round(digs, rnd);
        return *this;
    }
    bigdec& round()
    {
        if (!isfloat())
        {
            int digs = val.size() + user_scale - this_scale;
            if (digs > max_digs)
            {
                digs = max_digs;
            }
            do_round(digs, rnd);
        }
        return *this;
    }

//
// Comparisons:
//
private:
    using aligned_digits = std::pair<absval,absval>;
    aligned_digits align_digits(const bigdec& rhs) const;
    static int cmp(const absval&, const absval&) noexcept;
    static int raw_cmp(const absval&, const absval&);
    int cmp(const bigdec&) const;
public:
    bool operator==(const bigdec& rhs) const
    {
        return cmp(rhs) == 0;
    }
    bool operator!=(const bigdec& rhs) const
    {
        return cmp(rhs) != 0;
    }
    bool operator<(const bigdec& rhs) const
    {
        return cmp(rhs) < 0;
    }
    bool operator>(const bigdec& rhs) const
    {
        return cmp(rhs) > 0;
    }
    bool operator<=(const bigdec& rhs) const
    {
        return cmp(rhs) <= 0;
    }
    bool operator>=(const bigdec& rhs) const
    {
        return cmp(rhs) >= 0;
    }

//
// Arithmetic:
//
private:
    void do_inc_dec(bool);
public:
    bigdec& operator++()
    {
        do_inc_dec(false);
        return *this;
    }
    bigdec& operator--()
    {
        do_inc_dec(true);
        return *this;
    }

    bigdec operator++(int)
    {
        bigdec temp(*this);
        do_inc_dec(false);
        return temp;
    }
    bigdec operator--(int)
    {
        bigdec temp(*this);
        do_inc_dec(true);
        return temp;
    }

private:
    void add(absval&, const absval&);
    void sub(absval&, const absval&);
public:
    bigdec& operator+=(const bigdec&);
    bigdec& operator-=(const bigdec&);

private:
    void mul_part(absval&, elem_type);
    void mul(absval&, const absval&);
public:
    bigdec& operator*=(const bigdec&);

private:
    elem_type div_part(absval&, const absval&, elem_type, elem_type);
    int div(absval&, const absval&);
public:
    bigdec& operator/=(const bigdec&);

//
// size(), etc.
//
    bool empty() const noexcept
    {
        return val.empty();
    }
    std::size_t size() const noexcept
    {
        return val.size();
    }
    void shrink_to_fit()
    {
        val.shrink_to_fit();
    }

    std::size_t unnormalized_digits() const noexcept
    {
        return size();
    }
    std::size_t digits()
    {
        round();
        return unnormalized_digits();
    }
};

//
// Nonmember functions:
//

using std::swap;
inline void swap(bigdec& lhs, bigdec& rhs)
{
    lhs.swap(rhs);
}

inline std::string to_string(const bigdec& value,
                             bool scientific = false,
                             int precision = INT_MAX)
{
    return std::move(value.to_string(scientific, precision));
}

inline bigint to_bigint(const bigdec& val, rounding r = rounding::all_to_zero)
{
    return std::move(val.to_bigint(r));
}

//
// Arithmetic operators:
//

inline bigdec operator+(const bigdec& rhs)
{
    return rhs;
}
inline bigdec operator-(const bigdec& rhs)
{
    bigdec temp(rhs);
    temp.negate();
    return temp;
}

inline bigdec operator+(const bigdec& lhs, const bigdec& rhs)
{
    bigdec temp(lhs);
    temp += rhs;
    return temp;
}
inline bigdec operator-(const bigdec& lhs, const bigdec& rhs)
{
    bigdec temp(lhs);
    temp -= rhs;
    return temp;
}
inline bigdec operator*(const bigdec& lhs, const bigdec& rhs)
{
    bigdec temp(lhs);
    temp *= rhs;
    return temp;
}
inline bigdec operator/(const bigdec& lhs, const bigdec& rhs)
{
    bigdec temp(lhs);
    temp /= rhs;
    return temp;
}

//
// A few <cmath>-like functions:
//

inline bigdec abs(const bigdec& val)
{
    bigdec temp(val);
    temp.abs();
    return temp;
}

inline bigdec round(const bigdec& val, rounding r = rounding::tie_away_zero)
{
    bigdec temp(val);
    temp.set_rounding(r);
    temp.set_user_scale(0);
    temp.round();
    return temp;
}

inline bigdec ceil(const bigdec& val)
{
    return std::move(round(val, rounding::all_to_pos_inf));
}

inline bigdec floor(const bigdec& val)
{
    return std::move(round(val, rounding::all_to_neg_inf));
}

inline bigdec trunc(const bigdec& val)
{
    return std::move(round(val, rounding::all_to_zero));
}

bigdec rint(const bigdec&);

bigdec modf(const bigdec&, bigdec*);

bigdec remainder(const bigdec&, const bigdec&,
                 rounding = rounding::tie_to_even);

inline bigdec fmod(const bigdec& x, const bigdec& y)
{
    return std::move(remainder(x, y, rounding::all_to_zero));
}

inline bigdec sqr(const bigdec& val)
{
    bigdec temp(val);
    temp *= val;
    return temp;
}

bigdec copysign(const bigdec&, const bigdec&);

inline bigdec fma(const bigdec& multiplicand,
                  const bigdec& multiplier,
                  const bigdec& addend)
{
    bigdec temp(multiplicand);
    temp *= multiplier;
    temp += addend;
    return temp;
}

} // namespace dbacc

#ifndef BIGDEC_NO_IO_OPERATORS
  #include "bigdec_io.hpp"
#endif

#endif // BIGDEC_HPP_INCLUDED
