//
// bigint.cpp
//
// Bill Seymour, 2022-06-16
//
// Copyright Bill Seymour 2022.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This translation unit defines the non-inline functions
// for the bigint class defined in bigint.hpp.
//
// Note that the value is stored in little-endian fashion,
// so val.front() is the least significant element
// and val.back() is the most significant.
//

#ifdef NDEBUG
  #define BIGINT_NO_IO_OPERATORS
#endif
#include "bigint.hpp"

#ifndef NDEBUG
  #include <iomanip>
  namespace
  {
    using std::cerr;
    using std::hex;
    using std::dec;
  }
#endif

#include <stdexcept>
#include <algorithm> // reverse

#include <cstdlib>
#include <cctype>
#include <cerrno>
#include <cfenv>
#include <cassert>

namespace
{
    using std::size_t;
    using std::move;
}

namespace dbacc {

//
// Compare normalized absolute values:
//
int bigint::compare(const absval& lhs, const absval& rhs) noexcept
{
    assert(lhs.empty() || lhs.back() != 0);
    assert(rhs.empty() || rhs.back() != 0);

    size_t lsize = lhs.size(), rsize = rhs.size();
    if (lsize < rsize)
    {
        return -1;
    }
    if (lsize > rsize)
    {
        return +1;
    }
    // The sizes are equal.
    for (size_t idx = lsize; idx > 0; )
    {
        --idx;
        if (lhs[idx] < rhs[idx])
        {
            return -1;
        }
        if (lhs[idx] > rhs[idx])
        {
            return +1;
        }
    }
    return 0;
}

//
// Arithmetic helpers:
//
void bigint::add(absval& lhs, const absval& rhs)
{
    ext_type carry = 0;

    iter  lit = lhs.begin(),  lend = lhs.end();
    citer rit = rhs.cbegin(), rend = rhs.cend();
    for ( ; lit != lend && rit != rend; ++lit, ++rit)
    {
        carry += *lit;
        carry += *rit;
        *lit = val_type(carry & val_mask);
        carry >>= val_bits;
    }
    for ( ; carry != 0 && rit != rend; ++rit)
    {
        //
        // No more lhs elements, but there are still rhs elements,
        // and there was a carry.
        //
        carry += *rit;
        lhs.push_back(val_type(carry & val_mask));
        carry >>= val_bits;
    }
    for ( ; rit != rend; ++rit)
    {
        //
        // No more carries, but still more rhs elements.
        //
        lhs.push_back(*rit);
    }
    for ( ; carry != 0 && lit != lend; ++lit)
    {
        //
        // No more rhs elements, but there are still lhs elements,
        // and there was a carry.
        //
        carry += *lit;
        *lit = val_type(carry & val_mask);
        carry >>= val_bits;
    }
    if (carry != 0)
    {
        //
        // No more elements on either side, but we still have one last carry.
        //
        lhs.push_back(val_type(carry));
    }
}

//
// Subtraction has the preconditions that the minuend be not less than
// the subtrahend, and both have been normalized.
//
void bigint::sub(absval& lhs, const absval& rhs)
{
    assert(compare(lhs, rhs) >= 0);
    // and so lhs.size() >= rhs.size()

    using borrow_type = std::make_signed<ext_type>::type;
    borrow_type borrow = 0;

    iter  lit = lhs.begin(),  lend = lhs.end();
    citer rit = rhs.cbegin(), rend = rhs.cend();
    for ( ; rit != rend; ++lit, ++rit)
    {
        borrow = borrow_type(*lit) - borrow;
        if (borrow < *rit)
        {
            borrow += val_mod;
        }
        *lit = val_type((borrow - *rit) & val_mask);
        borrow >>= val_bits;
    }

    if (borrow != 0)
    {
        assert(borrow == 1 && lit != lend);
        for ( ; lit != lend; ++lit)
        {
            if (*lit > 0)
            {
                --*lit;
                #ifndef NDEBUG
                  --borrow;
                #endif
                break;
            }
            // else
            *lit = val_mask; // and borrow from next-MSD
        }
        assert(borrow == 0);
    }
    pop_leading_zeros(lhs);
}

//
// The following is also called from do_div() and operator unsigned long long():
//
unsigned long long bigint::make_ull(const absval& av) noexcept
{
    assert(fits_in_ull(av.size()));

    unsigned long long v = 0;

    for (criter it = av.crbegin(), end = av.crend(); it != end; ++it)
    {
        v = (v << val_bits) + *it;
    }

    return v;
}
//
// Multiply by a single-element multiplier.
//
void bigint::mul_part(absval& lhs, val_type rhs)
{
    ext_type carry = 0;

    for (iter it = lhs.begin(), end = lhs.end(); it != end; ++it)
    {
        carry += ext_type(*it) * rhs;
        *it = val_type(carry & val_mask);
        carry >>= val_bits;
    }
    if (carry != 0)
    {
        lhs.push_back(val_type(carry & val_mask));
    }
}
//
// Multiply by a multi-element multiplier.
//
void bigint::mul(bigint& lhs, const bigint& rhs)
{
    if (lhs.is_zero())
    {
        return;
    }
    if (rhs.is_zero())
    {
        lhs.clear();
        return;
    }
    if (rhs.is_one())
    {
        lhs.neg = lhs.neg != rhs.neg;
        return;
    }
    if (lhs.is_one())
    {
        lhs.val = rhs.val;
        lhs.neg = lhs.neg != rhs.neg;
        return;
    }

    //
    // If the product would fit in an unsigned long long:
    //
    if (fits_in_ull(lhs.val.size() + rhs.val.size()))
    {
        unsigned long long multiplicand = make_ull(lhs.val);
        lhs.neg = lhs.neg != rhs.neg;
        lhs.val.clear();
        lhs.init(multiplicand * make_ull(rhs.val));
        return;
    }

    //
    // Else we're stuck with generating partial products and summing them.
    //
    std::vector<absval> parts;

    //
    // Compute partial products for each rhs element, least significant first.
    //
    for (citer it = rhs.val.cbegin(), end = rhs.val.cend(); it != end; ++it)
    {
        static absval part;
        part = lhs.val;
        mul_part(part, *it);
        parts.push_back(part);
    }

    //
    // Sum the partial products.
    //
    lhs.clear();
    absval::size_type shift = 0;
    for (absval& part : parts)
    {
        //
        // For each partial product, "shift left" one additional val_type.
        //
        part.insert(part.begin(), shift, val_type(0));
        add(lhs.val, part);
        ++shift;
    }
    lhs.neg = lhs.neg != rhs.neg;
    lhs.normalize();
}

//
// Divide a partial dividend by a partial divisor
// doing a binary search for a trial quotient.
//
bigint::ext_type bigint::div_part(absval& lhs, const absval& rhs,
                                  ext_type min, ext_type max)
{
    if (max > min)
    {
        ext_type trial_quot = (max + min) / 2;
        absval prod_part(rhs);
        mul_part(prod_part, val_type(trial_quot));

        // If quot * divisor > dividend, quot needs to be smaller.
        if (compare(prod_part, lhs) > 0)
        {
            return div_part(lhs, rhs, min, trial_quot);
        }

        absval trial_rem = lhs;
        sub(trial_rem, prod_part);

        // If rem > divisor, quot needs to be bigger.
        if (compare(trial_rem, rhs) >= 0)
        {
            return div_part(lhs, rhs, trial_quot + 1, max);
        }

        // Got a quotient and a remainder.
        lhs = move(trial_rem);
        return trial_quot;
    }
    return 0;
}
//
// Divide lhs by rhs; put the quotient in *qptr and the remainder in *rptr.
//
void bigint::do_div(bigint* qptr, bigint* rptr,
                    const bigint& lhs, const bigint& rhs)
{
    assert(qptr != &lhs);
    assert(qptr != &rhs);
    assert(rptr != &lhs);
    assert(rptr != &rhs);
    assert(qptr != rptr);

    if (rhs.is_zero())
    {
        throw std::domain_error("bigint division by zero");
    }

    //
    // If the divisor is +/-1, the quotient is +/-lhs and the remainder is 0.
    //
    if (rhs.is_one())
    {
        if (qptr != nullptr)
        {
            qptr->val = lhs.val;
            qptr->neg = lhs.neg != rhs.neg;
        }
        if (rptr != nullptr)
        {
            rptr->clear();
        }
        return;
    }
    //
    // If abs(lhs) < abs(rhs), the quotient is 0 and the remainder is lhs.
    //
    int cmp = compare(lhs.val, rhs.val);
    if (cmp < 0)
    {
        if (qptr != nullptr)
        {
            qptr->clear();
        }
        if (rptr != nullptr)
        {
            *rptr = lhs;
        }
        return;
    }
    //
    // If they're equal, the quotient is +/-1 and the remainder is 0.
    //
    if (cmp == 0)
    {
        if (qptr != nullptr)
        {
            qptr->set_to_one(lhs.neg != rhs.neg);
        }
        if (rptr != nullptr)
        {
            rptr->clear();
        }
        return;
    }

    //
    // If abs(lhs) fits in an unsigned long long, then abs(rhs) does too
    // (because we don't get here unless abs(lhs) > abs(rhs)), so we can
    // do the division more efficiently.
    //
    if (fits_in_ull(lhs.val.size()))
    {
        unsigned long long dividend = make_ull(lhs.val);
        unsigned long long divisor  = make_ull(rhs.val);

        if (qptr != nullptr)
        {
            unsigned long long quot = dividend / divisor;
            qptr->val.clear();
            qptr->init(quot);
            qptr->neg = quot != 0 && lhs.neg != rhs.neg;
        }
        if (rptr != nullptr)
        {
            unsigned long long rem = dividend % divisor;
            rptr->val.clear();
            rptr->init(rem);
            rptr->neg = rem != 0 && lhs.neg;
        }
        return;
    }

    //
    // If none of the above, then we're stuck with the "long division" routine
    // that we learned in primary school.
    //

    absval quot, rem;

    for (size_t idx = lhs.val.size(); idx > 0; )
    {
        // rem doubles as partial dividend.
        val_type next = lhs.val[--idx];
        if (!rem.empty() || next != 0)
        {
            rem.insert(rem.begin(), 1, next);
        }

        if (compare(rem, rhs.val) < 0) // Partial quotient is zero.
        {
            quot.insert(quot.begin(), 1, 0);
            continue;
        }

        ext_type quot_part =
            div_part(rem, rhs.val, 0, val_mod);
        if (quot_part > val_mask)
        {
            quot.insert(quot.begin(), 1, val_type(quot_part >> val_bits));
        }
        quot.insert(quot.begin(), 1, val_type(quot_part  & val_mask));
    }

    if (qptr != nullptr)
    {
        pop_leading_zeros(quot);
        qptr->neg = lhs.neg != rhs.neg && !quot.empty();
        qptr->val = move(quot);
    }
    if (rptr != nullptr)
    {
        rptr->neg = lhs.neg && !rem.empty();
        rptr->val = move(rem);
    }
}

//
// The arithmetic operators:
//
bigint& bigint::operator+=(const bigint& rhs)
{
    if (rhs.is_zero())
    {
        return *this;
    }
    if (this->is_zero())
    {
        return *this = rhs;
    }
    if (rhs.neg == this->neg)
    {
        add(this->val, rhs.val);
        return *this;
    }

    //
    // Both are non-zero and the signs differ.  Subtract the smaller
    // from the larger and maybe change the sign.
    //
    if (compare(this->val, rhs.val) >= 0)
    {
        sub(this->val, rhs.val);
    }
    else
    {
        absval temp(rhs.val);
        sub(temp, this->val);
        this->val = move(temp);
        neg = !neg;
    }
    disallow_neg_zero();
    return *this;
}
bigint& bigint::operator-=(const bigint& rhs)
{
    if (rhs.is_zero())
    {
        return *this;
    }
    if (this->is_zero())
    {
        this->val = rhs.val;
        this->neg = !rhs.neg;
        return *this;
    }
    if (this->neg != rhs.neg)
    {
        add(this->val, rhs.val);
        return *this;
    }

    //
    // Both are non-zero and the signs are the same.  Subtract the smaller
    // from the larger and maybe change the sign.
    //
    if (compare(this->val, rhs.val) >= 0)
    {
        sub(this->val, rhs.val);
    }
    else
    {
        absval temp(rhs.val);
        sub(temp, this->val);
        this->val = move(temp);
        neg = !neg;
    }
    disallow_neg_zero();
    return *this;
}
bigint& bigint::operator*=(const bigint& rhs)
{
    mul(*this, rhs);
    return *this;
}
bigint& bigint::operator/=(const bigint& rhs)
{
    bigint temp;
    do_div(&temp, nullptr, *this, rhs);
    return *this = move(temp);
}
bigint& bigint::operator%=(const bigint& rhs)
{
    bigint temp;
    do_div(nullptr, &temp, *this, rhs);
    return *this = move(temp);
}

//
// Bitwise operations:
//
bigint& bigint::complement() noexcept
{
    for (val_type& v : val)
    {
        v = ~v;
    }
    normalize();
    return *this;
}
bigint& bigint::operator&=(const bigint& rhs) noexcept
{
    iter  lit = this->val.begin(), lend = this->val.end();
    citer rit = rhs.val.cbegin(),  rend = rhs.val.cend();
    for ( ; lit != lend && rit != rend; ++lit, ++rit)
    {
        *lit &= *rit;
    }
    if (lit != lend) // The rest should be 0s.
    {
        val.erase(lit, lend);
    }
    normalize(); // The loop could have given us most significant 0s.
    return *this;
}
bigint& bigint::operator^=(const bigint& rhs)
{
    iter  lit = this->val.begin(), lend = this->val.end();
    citer rit = rhs.val.cbegin(),  rend = rhs.val.cend();
    for ( ; lit != lend && rit != rend; ++lit, ++rit)
    {
        *lit ^= *rit;
    }
    if (rit != rend) // Just copy the rest.
    {
        val.insert(lend, rit, rend);
    }
    else // The loop could have given us most significant 0s.
    {
        normalize();
    }
    return *this;
}
bigint& bigint::operator|=(const bigint& rhs)
{
    iter  lit = this->val.begin(), lend = this->val.end();
    citer rit = rhs.val.cbegin(),  rend = rhs.val.cend();
    for ( ; lit != lend && rit != rend; ++lit, ++rit)
    {
        *lit |= *rit;
    }
    if (rit != rend) // Just copy the rest.
    {
        val.insert(lend, rit, rend);
    }
    // No need to normalize since the loop can only add 1 bits.
    return *this;
}

//
// Shift helpers:
//
void bigint::shift_left(unsigned cnt)
{
    absval::size_type whole_elems = cnt / val_bits; // Save for later.

    // Left-shift just cnt % val_bits for now:
    if ((cnt %= val_bits) != 0)
    {
        ext_type carry = 0;
        for (val_type& v : val)
        {
            carry |= ext_type(v) << cnt;
            v = val_type(carry & val_mask);
            carry >>= val_bits;
        }
        if (carry != 0)
        {
            val.push_back(val_type(carry));
        }
    }

    // Left-shift complete elements (shifts in zeros):
    val.insert(val.begin(), whole_elems, val_type(0));
}
void bigint::shift_right(unsigned cnt)
{
    if (cnt >= val_bits) // Shifting out at least one whole element:
    {
        absval::size_type whole_elems = cnt / val_bits;
        if (whole_elems >= val.size())
        {
            clear(); // Shifting out at least all currents bits.
            return;
        }

        // Shift out complete elements:
        val.erase(val.begin(), val.begin() + whole_elems);

        // Bit count remaining to be shifted:
        cnt %= val_bits;
    }

    if (cnt != 0)
    {
        // Bits remaining in less significant elements:
        ext_type carry = ext_type(val[0]) >> cnt;

        // We'll be shifting more significant elements to the left
        // by the val_bits'-complement number of bits from now on:
        cnt = val_bits - cnt;

        for (size_t idx = 1, end = val.size(); idx != end; ++idx)
        {
            carry |= ext_type(val[idx]) << cnt;
            val[idx - 1] = carry & val_mask; // NB, set previous element.
            carry >>= val_bits;
        }

        // The most significant element:
        if (carry != 0)
        {
            val.back() = val_type(carry);
        }
        else // We shifted out the whole MSD.
        {
            val.pop_back();
            if (val.empty())
            {
                neg = false;
            }
        }
    }
}

//
// Shift operators:
//
bigint& bigint::operator<<=(int cnt)
{
    if (!is_zero())
    {
        if (cnt > 0)
        {
            shift_left(cnt);
        }
      // We probably have 2's-comp. ints, ...
      #if INT_MIN != -INT_MAX
        // so just in case ...
        else if (INT_MIN == cnt)
        {
            shift_right(unsigned(INT_MAX) + 1U);
        }
      #endif
        else if (cnt < 0)
        {
            shift_right(-cnt);
        }
    }
    return *this;
}
bigint& bigint::operator>>=(int cnt)
{
    if (!is_zero())
    {
        if (cnt > 0)
        {
            shift_right(cnt);
        }
      #if INT_MIN != -INT_MAX
        else if (INT_MIN == cnt)
        {
            shift_left(unsigned(INT_MAX) + 1U);
        }
      #endif
        else if (cnt < 0)
        {
            shift_left(-cnt);
        }
    }
    return *this;
}

//
// The constructor helpers:
//
void bigint::init(unsigned long long ull)
{
    assert(val.empty());
    // additional precondition:  neg set elsewhere
    for ( ; ull != 0; ull >>= val_bits)
    {
        val.push_back(val_type(ull & val_mask));
    }
}
void bigint::init(long long sll)
{
    neg = sll < 0;
    if (neg && (sll != LLONG_MIN || LLONG_MIN == -LLONG_MAX))
    {
        sll = -sll;
    }
    init(static_cast<unsigned long long>(sll));
}
void bigint::init(long double ld)
{
    assert(val.empty());

    if (!std::isfinite(ld))
    {
        throw std::invalid_argument(
            "Can't convert NaNs or infinities to bigint");
    }

    ld = std::trunc(ld);

    neg = ld < 0.0;
    if (neg)
    {
        ld = -ld;
    }

    static constexpr long double factor = val_mod;

    while (ld >= 1.0)
    {
        long double newval = std::trunc(ld / factor);
        long double newelem =
          #ifdef FP_FAST_FMAL
            std::fmal(-newval, factor, ld);
          #else
            ld - newval * factor;
          #endif
        val.push_back(val_type(newelem));
        ld = newval;
    }
}
void bigint::init(const char* s, int radix, const char** end)
{
    assert(val.empty());

    if (nullptr == s)
    {
        throw std::invalid_argument("Conversion from null const char*");
    }

    neg = false;
    if (*s == '-')
    {
        neg = true;
        ++s;
    }
    else if (*s == '+')
    {
        ++s;
    }

    if (radix == 16)
    {
        if (*s == '0')
        {
            ++s;
            if (*s == 'x' || *s == 'X')
            {
                ++s;
            }
        }
    }
    else if (radix != 10 && radix != 8) // infer radix from string
    {
        if (*s == '0')
        {
            ++s;
            if (*s == 'x' || *s == 'X')
            {
                radix = 16;
                ++s;
            }
            else
            {
                radix = 8;
            }
        }
        else
        {
            radix = 10;
        }
    }

    for ( ; *s != '\0'; ++s)
    {
        unsigned char c = *s;

        bool valid_digit =
            std::isdigit(c) && (radix != 8 || c < '8') ||
            radix == 16 && std::isxdigit(c);

        if (!valid_digit)
        {
            break; // OK, just terminate the parse.
        }

        if (std::isdigit(c))
        {
            c -= '0';
        }
        else if (std::isupper(c))
        {
            c -= 'A' - 10;
        }
        else
        {
            c -= 'a' - 10;
        }

        ext_type carry = c;
        for (val_type& v : val)
        {
            carry += ext_type(v) * radix;
            v = val_type(carry & val_mask);
            carry >>= val_bits;
        }
        if (carry != 0)
        {
            val.push_back(val_type(carry));
        }
    }

    if (end != nullptr)
    {
        *end = s;
    }

    disallow_neg_zero();
}

std::string bigint::to_string(int radix) const
{
    assert(val.empty() || val.back() != 0);

    std::string retval;

    if (is_zero())
    {
        retval.assign(1, '0');
    }
    else
    {
        if (radix != 8 && radix != 16)
        {
            radix = 10;
        }
        const bigint divisor(radix);

        bigint dividend(this->val), quot, rem;

        // the digits in reverse order:
        do {
            do_div(&quot, &rem, dividend, divisor);
            char c = rem.empty() ? '0' : char(rem.val.back() + '0');
            if (c > '9')
            {
                c += 'A' - ('9' + 1);
            }
            retval.push_back(c);
            dividend.val = move(quot.val);
        } while (!dividend.empty());

        while (!retval.empty() && '0' == retval.back())
        {
            retval.pop_back();
        }
        assert(!retval.empty());

        if (neg)
        {
            retval.push_back('-');
        }

        std::reverse(retval.begin(), retval.end());
    }

    return retval;
}

//
// Do ++ or --:
//
void bigint::do_inc_dec(bool n)
{
    if (val.empty())
    {
        set_to_one(n);
    }
    else
    {
        static const absval one(1, 1);
        if (n == neg)
        {
            add(val, one);
        }
        else
        {
            sub(val, one); // already pops leading zeros
            disallow_neg_zero();
        }
    }
}

long double bigint::make_ld() const noexcept
{
  #pragma STDC FENV_ACCESS ON

    static constexpr long double factor = val_mod;

    long double retval = 0.0;

    for (criter it = val.crbegin(), end = val.crend(); it != end; ++it)
    {
        std::feclearexcept(FE_ALL_EXCEPT);
        retval =
          #ifdef FP_FAST_FMAL
            std::fmal(retval, factor, static_cast<long double>(*it));
          #else
            retval * factor + *it;
          #endif
        if (std::fetestexcept(FE_OVERFLOW))
        {
            if (std::numeric_limits<long double>::has_quiet_NaN)
            {
                return std::numeric_limits<long double>::quiet_NaN();
            }
            if (std::numeric_limits<long double>::has_infinity)
            {
                return std::numeric_limits<long double>::infinity();
            }
            errno = ERANGE;
            return HUGE_VALL;
        }
    }

    return neg && retval != 0.0 ? -retval : retval;
}

int bigint::compare(const bigint& rhs) const noexcept
{
    if (this->neg && !rhs.neg)
        return -1;
    if (!this->neg && rhs.neg)
        return +1;
    int cmp = compare(this->val, rhs.val);
    return neg && cmp != 0 ? -cmp : cmp;
}

//
// The two non-inline non-member functions:
//
bigint gcd(const bigint& arg1, const bigint& arg2)
{
    if (!arg1.is_zero() || !arg2.is_zero()) // as expected
    {
        bigint a(arg1), c(arg2); // Start with non-const copies.
        a.abs();
        c.abs();

        //
        // After Euclid (Knuth, vol 2, sect 4.5.2):
        //
        for (bigint r; !c.is_zero(); )
        {
            r = move(a);
            r %= c;
            a = move(c);
            c = move(r);
        }
        return a;
    }
    throw std::invalid_argument("dbacc::gcd(0,0)");
}
bigint lcm(const bigint& a, const bigint& c)
{
    if (!a.is_zero() && !c.is_zero()) // as expected
    {
        bigint retval(a);
        retval /= gcd(a,c);
        retval *= c;
        retval.abs();
        return retval;
    }
    throw std::invalid_argument("dbacc::lcm() has a zero argument");
}

} // namespace dbacc

// End of bigint.cpp
